@@ -1,5 +1,6 @@
 {-# LANGUAGE MultiParamTypeClasses, TypeOperators
-           , TypeFamilies, UndecidableInstances
+           , TypeFamilies, UndecidableInstances, CPP
+           , FlexibleContexts
  #-}
 {-# OPTIONS_GHC -Wall #-}
 ----------------------------------------------------------------------
@@ -32,6 +33,7 @@ module Data.VectorSpace
 
 import Control.Applicative (liftA2)
 import Data.Complex hiding (magnitude)
+import Foreign.C.Types (CSChar, CInt, CShort, CLong, CLLong, CIntMax, CFloat, CDouble)
 import Data.Ratio
 
 import Data.AdditiveGroup
@@ -48,7 +50,7 @@ class AdditiveGroup v => VectorSpace v where
 infixr 7 <.>
 
 -- | Adds inner (dot) products.
-class VectorSpace v => InnerSpace v where
+class (VectorSpace v, AdditiveGroup (Scalar v)) => InnerSpace v where
   -- | Inner/dot product
   (<.>) :: v -> v -> Scalar v
 
@@ -84,19 +86,27 @@ normalized :: (InnerSpace v, s ~ Scalar v, Floating s) =>  v -> v
 normalized v = v ^/ magnitude v
 
 -- | @project u v@ computes the projection of @v@ onto @u@.
-project :: (InnerSpace v, s ~ Scalar v, Floating s) => v -> v -> v
-project u v = (v <.> u') *^ u'
-  where u' = normalized u
+project :: (InnerSpace v, s ~ Scalar v, Fractional s) => v -> v -> v
+project u v = ((v <.> u) / magnitudeSq u) *^ u
 
-instance VectorSpace Double where
-  type Scalar Double = Double
-  (*^) = (*)
-instance InnerSpace  Double where (<.>) = (*)
+#define ScalarType(t) \
+  instance VectorSpace (t) where \
+    { type Scalar (t) = (t) \
+    ; (*^) = (*) } ; \
+  instance InnerSpace  (t) where (<.>) = (*)
 
-instance VectorSpace Float  where
-  type Scalar Float = Float
-  (*^)  = (*)
-instance InnerSpace  Float  where (<.>) = (*)
+ScalarType(Int)
+ScalarType(Integer)
+ScalarType(Double)
+ScalarType(Float)
+ScalarType(CSChar)
+ScalarType(CInt)
+ScalarType(CShort)
+ScalarType(CLong)
+ScalarType(CLLong)
+ScalarType(CIntMax)
+ScalarType(CDouble)
+ScalarType(CFloat)
 
 instance Integral a => VectorSpace (Ratio a) where
   type Scalar (Ratio a) = Ratio a
@@ -107,7 +117,7 @@ instance (RealFloat v, VectorSpace v) => VectorSpace (Complex v) where
   type Scalar (Complex v) = Scalar v
   s*^(u :+ v) = s*^u :+ s*^v
 
-instance (RealFloat v, InnerSpace v, s ~ Scalar v, AdditiveGroup s)
+instance (RealFloat v, InnerSpace v)
      => InnerSpace (Complex v) where
   (u :+ v) <.> (u' :+ v') = (u <.> u') ^+^ (v <.> v')
 
@@ -127,8 +137,7 @@ instance ( VectorSpace u, s ~ Scalar u
   s *^ (u,v) = (s*^u,s*^v)
 
 instance ( InnerSpace u, s ~ Scalar u
-         , InnerSpace v, s ~ Scalar v
-         , AdditiveGroup (Scalar v) )
+         , InnerSpace v, s ~ Scalar v )
     => InnerSpace (u,v) where
   (u,v) <.> (u',v') = (u <.> u') ^+^ (v <.> v')
 
@@ -141,8 +150,7 @@ instance ( VectorSpace u, s ~ Scalar u
 
 instance ( InnerSpace u, s ~ Scalar u
          , InnerSpace v, s ~ Scalar v
-         , InnerSpace w, s ~ Scalar w
-         , AdditiveGroup s )
+         , InnerSpace w, s ~ Scalar w )
     => InnerSpace (u,v,w) where
   (u,v,w) <.> (u',v',w') = u<.>u' ^+^ v<.>v' ^+^ w<.>w'
 
@@ -157,8 +165,7 @@ instance ( VectorSpace u, s ~ Scalar u
 instance ( InnerSpace u, s ~ Scalar u
          , InnerSpace v, s ~ Scalar v
          , InnerSpace w, s ~ Scalar w
-         , InnerSpace x, s ~ Scalar x
-         , AdditiveGroup s )
+         , InnerSpace x, s ~ Scalar x )
     => InnerSpace (u,v,w,x) where
   (u,v,w,x) <.> (u',v',w',x') = u<.>u' ^+^ v<.>v' ^+^ w<.>w' ^+^ x<.>x'
 
@@ -196,7 +203,7 @@ instance (HasTrie a, VectorSpace v) => VectorSpace (a :->: v) where
 
 
 
-instance (InnerSpace a, AdditiveGroup (Scalar a)) => InnerSpace (Maybe a) where
+instance InnerSpace a => InnerSpace (Maybe a) where
   -- dotting with zero (vector) yields zero (scalar)
   Nothing <.> _     = zeroV
   _ <.> Nothing     = zeroV
